from . import bibtex
from . import journals
from . import recognise
from . import reformat
from . import tex
from ._version import version
from ._version import version_tuple
from .journals import get_configdir
